#include <unistd.h>
#include <stdarg.h>
#include <View.h>
#include <Bitmap.h>

#include "maze.h"
#include "mazefield.h"
#include "graphics.h"

static long entry_func( void * );

static long entry_func( void *arg ){
	MazeField *obj = (MazeField *)arg;
	obj->construct();
	return 0;
}

MazeField::MazeField( MazeWindow *win, BRect rect, long flag )
          :BView( rect, "", flag, B_WILL_DRAW | B_PULSE_NEEDED ){
	BRect bitmap_rect;

	window = win;
	bitmap_rect.Set( 0, 0, FIELD_WIDTH-1, FIELD_HEIGHT-1 );
	field = new MazeBitmap( bitmap_rect, 0x00 );
}

bool MazeField::make( int x, int y ){    // private method
	struct cst {
		Maze *ptr;
		int dirhist;
	} *construction;
	int constructionlength;
	int i, nowx, nowy, newx, newy, nowlength, direction;
	struct mawari *syuui;

	if( maze[x][y]->IsFree() == FALSE ){ return TRUE; }
	nowx = x;
	nowy = y;
	nowlength = 0;
	constructionlength = CONSTRUCTIONUNIT;
	construction = (struct cst *)malloc( sizeof(*construction)*constructionlength );
	while(1){
		syuui = maze[nowx][nowy]->search( FREE | WALL );
		if( syuui->nOKs != 0 ){  // nobasu
			direction = syuui->direction[rand() % syuui->nOKs];
			switch( direction ){
				case UP:
					newx = nowx; newy = nowy-1;
					break;
				case DOWN:
					newx = nowx; newy = nowy+1;
					break;
				case LEFT:
					newx = nowx-1; newy = nowy;
					break;
				case RIGHT:
					newx = nowx+1; newy = nowy;
					break;
			}
			construction[nowlength].dirhist = direction;
			if( maze[newx][newy]->IsFree() == FALSE ){  // osimai
				maze[nowx][nowy]->connectwith( direction, CONNECT | DRAW, this );
				for( i = 0; i < nowlength; i++ ){
					construction[i].ptr->SetUnderConstruction( FALSE );
				}
				return TRUE;
			}else{  // nobasu
				maze[nowx][nowy]->connectwith( direction, CONNECT | DRAW, this );
				construction[nowlength].ptr = maze[nowx][nowy];
				construction[nowlength].ptr->SetUnderConstruction( TRUE );
				nowx = newx;
				nowy = newy;
			}
		}else{ // jibun to butukatta
			for( i = nowlength-1; i >= 0; i-- ){
				construction[i].ptr->connectwith( construction[i].dirhist, DISCONNECT | DRAW, this );
				construction[i].ptr->SetUnderConstruction( FALSE );
			}
			nowlength = 0;
			return FALSE;
		}
		nowlength++;
		if( nowlength >= constructionlength ){
			constructionlength += CONSTRUCTIONUNIT;
			construction = (struct cst *)realloc( construction, sizeof(*construction)*constructionlength );
		}
	}
}

long MazeField::Go(){  // see The Kernel Kit p.8
	my_thread = spawn_thread( entry_func, "maze", B_NORMAL_PRIORITY, this );
	return resume_thread( my_thread );
}

// Draw version of easy to call
void MazeField::Draw(){
	window->Lock();
	DrawBitmap( field, BPoint( 0, 0 ) );
	window->Unlock();
}
// Draw version of faster  bug bug bug bug
void MazeField::Draw( BPoint siten, BPoint syuten ){
	window->Lock();
	DrawBitmapAsync( field, BRect( siten, syuten ) );
	window->Unlock();
}
// Draw version for BLooper
void MazeField::Draw( BRect aRect ){
	Draw();
}

void MazeField::Pulse(){
	Draw();
}

void MazeField::InitField(){
	int x, y;

	window->Lock();
	for( y = 0; y < MAZE_Y; y++ ){
		for( x = 0; x < MAZE_X; x++ ){
			maze[x][y] = new Maze( field, x, y );
		}
	}
	for( y = 0; y < MAZE_Y; y++ ){
		for( x = 0; x < MAZE_X; x++ ){
			maze[x][y]->setneighbor(
				DIRECTIONS,
				y-1 >= 0     ? maze[x][y-1] : NULL,
				y+1 < MAZE_Y ? maze[x][y+1] : NULL,
				x-1 >= 0     ? maze[x-1][y] : NULL,
				x-1 < MAZE_X ? maze[x+1][y] : NULL
			);
		}
	}
	window->Unlock();
}

void MazeField::construct(){
	int x, y, issucceeded;

	for( x = 0, y = 0; x < MAZE_X-1; x++ ){
		maze[x][y]->connectwith( RIGHT, CONNECT | DRAW, this );
	}
	for( x = MAZE_X-1, y = 0; y < MAZE_Y-1; y++ ){
		maze[x][y]->connectwith( DOWN, CONNECT | DRAW, this );
	}
	for( x = MAZE_X-1, y = MAZE_Y-1; x > 0; x-- ){
		maze[x][y]->connectwith( LEFT, CONNECT | DRAW, this );
	}
	for( x = 0, y = MAZE_Y-1; y > 0; y-- ){
		maze[x][y]->connectwith( UP, CONNECT | DRAW, this );
	}
	maze[0][MAZE_Y/2]->connectwith( UP, DISCONNECT | DRAW, this );
	maze[MAZE_X-1][MAZE_Y/2]->connectwith( UP, DISCONNECT | DRAW, this );

	for( y = 1; y < MAZE_Y-1; y++ ){
		for( x = 1; x < MAZE_X-1; x++ ){
			do{
				issucceeded = make( x, y );
			}while( issucceeded != TRUE );
		}
	}	
}

Maze::Maze( MazeBitmap *fld, int xx, int yy ){
	int i;
	BPoint point;

	field = fld;
	for( i = 0; i < DIRECTIONS; i++ ){
		connection[i] = FALSE;
	}

	x = xx;
	y = yy;
	point.Set( LEFTSKIP + x * PITCH_X,
	           UPPERSKIP + y * PITCH_Y );
	field->Pset( point, 0xff );

	SetUnderConstruction( FALSE );
}

void Maze::setneighbor( int n, ... ){
	va_list np;
	int i;

	va_start( np, n );
	for( i = 0; i < n; i++ ){
		neighbor[i] = va_arg( np, Maze * );
	}
	va_end( np );
}

void Maze::connectwith( int houkou, int mode, MazeField *mazefield ){
	int isconnect = mode & CONNECT;
	int xtonari, ytonari;
	BPoint siten, syuten;

	if( neighbor[houkou] == NULL ){  return; }        // must be error!
	connection[houkou] = isconnect == 0 ? FALSE : TRUE;
	if( (mode & DRAW) == SETONLY ){ return; }
	xtonari = houkou == RIGHT ?  1 :
	          houkou == LEFT  ? -1 : 0;
	ytonari = houkou == UP   ? -1 :
	          houkou == DOWN ?  1 : 0;
	neighbor[houkou]->connectwith( reverse(houkou), isconnect | SETONLY, mazefield );
	siten  = BPoint( LEFTSKIP  +       x       * PITCH_X + xtonari,
	                 UPPERSKIP +       y       * PITCH_Y + ytonari );
	syuten = BPoint( LEFTSKIP  + (x + xtonari) * PITCH_X - xtonari,
	                 UPPERSKIP + (y + ytonari) * PITCH_Y - ytonari );
	field->Line( siten, syuten, isconnect == 1 ? 0x44 : 0x00 );
}

struct mawari *Maze::search( int conditions ){
	static struct mawari answer;
	int i;

	answer.nOKs = 0;
	for( i = 0; i < DIRECTIONS; i++ ){
		if((conditions & FREE) == FREE &&  // search free
		   (neighbor[i]->IsFree() == TRUE)){
			answer.direction[answer.nOKs++] = i;
		}		    
		if((conditions & WALL) == WALL &&  // search wall
		   (neighbor[i]->IsFree() == FALSE) &&
		   (neighbor[i]->IsUnderConstruction() == FALSE)){
			answer.direction[answer.nOKs++] = i;
		}
		if((conditions & SELF) == SELF &&  // search the part of self
		   (neighbor[i]->IsFree() == FALSE) &&
		   (neighbor[i]->IsUnderConstruction() == TRUE)){
			answer.direction[answer.nOKs++] = i;
		}
	}
	return &answer;
}

bool Maze::IsFree(){
	int i;
	bool rts = TRUE;

	for( i = 0; i < DIRECTIONS; i++ ){
		if( connection[i] == TRUE ){
			rts = FALSE;
			break;
		}
	}
	return rts;
}

bool Maze::SetUnderConstruction( bool suc ){
	under_construction = suc;
}

bool Maze::IsUnderConstruction(){
	return under_construction;
}

int Maze::reverse( int direction ){
	switch( direction ){
		case UP:
			return DOWN;
		case DOWN:
			return UP;
		case LEFT:
			return RIGHT;
		case RIGHT:
			return LEFT;
	}
}
